#![feature(const_fn)]

use core::cell::UnsafeCell;
use core::mem::transmute;

pub struct Handler {
    // Handler that will be executed on `call`.
    h: UnsafeCell<*const dyn FnMut()>,
}

impl Handler {
    // Returns a new Handler that initially does nothing when
    // called. Override it's behavior by using `replace`.
    pub const fn new() -> Self {
        Self {
            h: UnsafeCell::new(&Self::default_handler),
        }
    }

    // Replace the behavior of this handler with `f`.
    //
    // This function transmutes `f` into a static lifetime, which it
    // probably isn't. You need to make sure that either `f` actually
    // lives forever or you need to replace it before it goes out of
    // scope.
    //
    // There is no exclusion on replacing the handler's behavior while
    // it is being executed. It is your responsiibility to make sure
    // that it's not being executed when you call `replace`.
    pub unsafe fn replace(&self, f: &dyn FnMut()) {
        let f_static: &'static dyn FnMut() = transmute(f);
        *self.h.get() = f_static;
    }

    // Execute this handler.
    //
    // This function assumes that a replace is not occurring when the
    // closure is being looked up. You need to ensure that `replace`
    // and `call` can not occur at the same time.
    pub unsafe fn call(&self) {
        let f: &mut dyn FnMut() = &mut *(*self.h.get() as *mut dyn FnMut());
        f();
    }

    // Do nothing handler. Needed by `call` until `replace` is used to
    // set specific behavior.
    fn default_handler() {}
}

unsafe impl Sync for Handler {}

static mut HANDLER: Handler = Handler::new();
